unit KOLXMLParser;
{ 
Simple XML parser skeleton object for KOL Library
01.09.2011
rednakse
}

{$DEFINE Atr_EX}
{$DEFINE REPLACE_1310}
//{$DEFINE CP_UTILS}
interface

uses Windows, KOL{$IFDEF CP_UTILS}, cputils{$ENDIF}, Err;

type
  PCharBuffer = ^TCharBuffer;
  PXMLAtribute = ^TXMLAtribute;
  PXMLAtributeList = ^TXMLAtributeList;
  PXMLTokenizer = ^TXMLTokenizer;
  PXMLNode = ^TXMLNode;
  PXMLParser = ^TXMLParser;

  PBytes = ^TBytes;
  TBytes = array of byte;

  TEncoding = (encUCS2LE, encUCS2BE, encUTF8, encANSI);
  TTokenObj = (tEOF, tOpen, tClose, tOpenClose, tHeader, tComment, tEmpty, tCDATA, tText);
  TStage = (stgRead, stgDecode, stgParse);
  
  TXMLSaveLoad = procedure(Sender: PXMLParser; Stage: TStage; Token: TTokenObj; NamedIteam: PKOLChar; Readed, Total: cardinal) of object;

  TCharBuffer = object
  {       }
  private
    FBuf: array of KOLChar;
    FSize: cardinal;
    FPos : cardinal;
    function FGetSize: cardinal;
    procedure FAlloc(l: cardinal);
  public
    Memory: Pointer;
    property Count: cardinal read FPos;
    property Size: cardinal read FGetSize;
    procedure PutStr(const S: KOLString);
    procedure PutChar(const Ch: KOLChar);
    procedure PutBuf(const P: PKOLChar);
    procedure WrtStr(var S: KOLString);
    procedure WrtStream(P: PStream; CP: TEncoding);
    function GetStr: KOLString;
    procedure Clear;
    constructor Create;
    destructor Destroy;
   end;
   
  TXMLAtribute = packed record
  { ''  }
    AT_Name,
    AT_Value: KOLstring;
  end;
  
  TXMLAtributeList = object(TObj)
  {      }
  private
    FAtrList : PList;
    function FGetAtributeByIndex(index: integer): PXMLAtribute;
    function FGetAtributeName(index: integer): KOLString;
    procedure FSetAtributeName(index: integer; const value: KOLString);
    function FGetAtributeValue(index: integer): KOLString;
    procedure FSetAtributeValue(index: integer; const value: KOLString);
    function FGetAtribute(AtrName: KOLString): KOLString;
    procedure FSetAtribute(AtrName: KOLString; const value: KOLString);
    function FGetCount: integer;
    {$IFDEF Atr_EX}
    function FGetAsInteger(AtrName: KOLString): integer;
    function FGetAsBoolean(AtrName: KOLString): boolean;
    function FGetAsFloat(AtrName: KOLString): extended;
    function FGetAsDateTime(AtrName: KOLString): TDateTime;
    procedure FSetAsInteger(AtrName: KOLString; const value: integer);
    procedure FSetAsBoolean(AtrName: KOLString; const value: boolean);
    procedure FSetAsFloat(AtrName: KOLString; const value: extended);
    procedure FSetAsDateTime(AtrName: KOLString; const value: TDateTime);
    {$ENDIF}
    function FGetLine: KOLString;
    procedure FSetLine(const value: KOLString);
  public
    property Atribute[AtrName: KOLString]: KOLString read FGetAtribute write FSetAtribute;
    { RW       }
    {$IFDEF Atr_EX}
    { ,       }
    property AsInteger[AtrName: KOLString]: integer read FGetAsInteger write FSetAsInteger;
    property AsBoolean[AtrName: KOLString]: boolean read FGetAsBoolean write FSetAsBoolean;
    property AsFloat[AtrName: KOLString]: extended read FGetAsFloat write FSetAsFloat;
    property AsDateTime[AtrName: KOLString]: TDateTime read FGetAsDateTime write FSetAsDateTime;
    {$ENDIF}
    function IndexByName(AtrName: KOLString): integer;
    {      ;   >= 0,  -1 }
    property AtributeName[index: integer] : KOLString read FGetAtributeName write FSetAtributeName;
    { RW      }
    property AtributeValue[index: integer] : KOLString read FGetAtributeValue write FSetAtributeValue;
    { RW      }
    property AtributeByIndex[index: integer]: PXMLAtribute read FGetAtributeByIndex;
    { RO        }
    property Line: KOLString read FGetLine write FSetLine;
    { RW        }
    procedure Remove(index: integer);
    {     }
    procedure Move(CurIdx, NewIdx: integer);
    {    }
    property Count: integer read FGetCount;
    {   }
    procedure Clear;
    {   }
    constructor Create;
    destructor Destroy; virtual;
  end;
  
  TXMLTokenizer = object(TObj)
  {   }
  private
    FBuf    : PKOLChar;
    FToken  : TTokenObj;
    FCnt    : cardinal;
    FLen    : cardinal;
    FBBuf   : PCharBuffer;
    FBody   : KOLString;
    procedure FSkipBlanks;
    function FReadCom: TTokenObj;
    function FReadText: TTokenObj;
    function FReadComment: TTokenObj;
    function FReadCDATA: TTokenObj;
    function Chk_Header: boolean;
    function Chk_Comment: boolean;
    function Chk_Close: boolean;
    function Chk_OpenClose: boolean;
    function Chk_CDATA: boolean;
    //function FGetBody: KOLString;
  public
    procedure Clear;
    {  ,   }
    procedure Initilize(const Buf: PKOLChar; const Len: cardinal);
    {  , 
      Buf   XML   Len  }
    property Body: KOLString read FBody;
    {    }
    property Token: TTokenObj read FToken;
    {     }
    function NextToken: TTokenObj;
    {      }
    constructor Create;
    destructor Destroy; virtual;
  end;
  
  TXMLNode = object(TObj)
  {   XML- }
  private
    FAtributes: PXMLAtributeList;
    FRoot,
    FParent: PXMLNode;
    FNodeList: PList;
    FParentCount: cardinal;
    FNDName,
    FText: KOLstring;
    FParser: PXMLParser;
    FAssigned: boolean;
    FTokenizer: PXMLTokenizer;
    FOnLoad : TXMLSaveLoad;
    procedure FSetName(const value: KOLString);
    procedure FSetText(const value: KOLString);
    //[!nontested]
    function FGetXML: KOLString;
    //[!nontested]
    procedure FSetXML(const value: KOLString);
    function FGetID: KOLString;
    procedure FSetID(const value: KOLString);
    //[!nonended]
    procedure FSetParent(const value: PXMLNode);
    function FGetFirstNode: PXMLNode;
    function FGetLastNode: PXMLNode;
    function FGetChildCount: integer;
    function FGetNodeByIndex(index: integer): PXMLNode;
    function FGetNodeByName(NodeName: KOLString): PXMLNode;
    function FGetNodeIndex(NodeName: KOLString): integer;
    function FGetNodeById(ID_String: KOLString): PXMLNode;
    function FCheckHasAtributes: boolean;
    function FCheckBlanked: boolean;
    function FCheckChildrens: boolean;
    function FCheckText: boolean;
    function FCheckSafeAccess: boolean;
    function FCheckFormated: boolean;
    function FGetEncoding: TEncoding;
    procedure FSaveToStream(Stream: PStream);
    function FGetTabulate: KOLstring;
    function FGetOpenToken: KOLstring;
    function FGetOpenCloseToken: KOLstring;
    function FGetCloseToken: KOLstring;
    procedure FLoadFromToken;
  public
    //[!nontested]
    property XML: KOLString read FGetXML write FSetXML;
    {     XML }
    property ND_Name: KOLString read FNDName write FSetName;
    {   }
    property ND_Text: KOLString read FText write FSetText;
    {   }
    property ND_ID: KOLString read FGetID write FSetID;
    { ID  (   ID   ID_ID) }    
    property Atributes: PXMLAtributeList read FAtributes;
    {   }
    property Root: PXMLNode read FRoot;
    { RO   }
    property Parent: PXMLNode read FParent write FSetParent;
    { RW   }
    property FirstNode: PXMLNode read FGetFirstNode;
    { RO     }
    property LastNode: PXMLNode read FGetLastNode;
    { RO    ,  NodeCount = 1   FirstNode }
    function NewNode: PXMLNode;
    {    }
    property NodeCount: integer read FGetChildCount;
    { RO    }
    property Node[index: integer]: PXMLNode read FGetNodeByIndex;
    {      , SafeAccess  }
    property NodeByName[NodeName: KOLString]: PXMLNode read FGetNodeByName;
    {       }
    function NodeIndexFrom(NodeName: KOLString; IndexFrom: integer = 0): integer;
    {     }
    function FindNodeByAtribute(AtrName, Str: KOLString; IndexFrom: integer = 0): PXMLNode;
    {     }
    property GetNodeIndex[NodeName: KOLString]: integer read FGetNodeIndex;
    {      }
    property GetNodeById[ID_string: KOLString]: PXMLNode read FGetNodeById;
    {    ID }
    procedure MoveNode(CurIdx, NewIdx: integer);
    {   }
    procedure RemoveNode(index: cardinal);
    {     }
    procedure Clear;
    {   ,  Assigned }
    property IsAssigned: boolean read FAssigned;
    { RO    (  ,  ,   ) }
    property HasAtributes: boolean read FCheckHasAtributes;
    { RO   }
    property Blanked: boolean read FCheckBlanked;
    { RO  , HasChild = HasText = FALSE }
    property HasChild: boolean read FCheckChildrens;
    { RO    }
    property HasText: boolean read FCheckText;
    { RO   }
    property SafeAccess: boolean read FCheckSafeAccess;
    { RO     ,       , 
           -   }
    property Formated: boolean read FCheckFormated;
    { RO  / ( ) }
    property Encoding: TEncoding read FGetEncoding;
    { RO  ( ) }
    //[!nontested]
    property OnLoad: TXMLSaveLoad read FOnLoad write FOnLoad;
    {        ,   ..
               }
    //[!nontested]
    procedure Include ( P: PXMLNode);
    {       ,    }
    constructor Create;
    destructor Destroy; virtual;
  end;

  TXMLParser = object(TObj)
  private
    FRoot: PXMLNode;
    FSafeAccess: boolean;
    FFormated: boolean;
    FHasHeader: boolean;
    FHeader : PXMLAtributeList;
    FToken: PXMLTokenizer;
    FEncoding: TEncoding;
    FOnLoad,
    FOnSave : TXMLSaveLoad;
    procedure FSetEncoding(const value: TEncoding);
  public
    property Header: PXMLAtributeList read FHeader;
    {  XML- ( ) }
    property Root: PXMLNode read FRoot;
    {    ,  HasHeader = false     (!) }
    property Encoding: TEncoding read FEncoding write FSetEncoding;
    {  }
    property SafeAccess: boolean read FSafeAccess write FSafeAccess;
    {    (. SafeAccess of Node) }
    property Formated: boolean read FFormated write FFormated;
    { \  (, CRLF) }
    property HasHeader: boolean read FHasHeader write FHasHeader;
    {   , .  Root of Parser }
    procedure LoadFromStream(Stream: PStream);
    function LoadFromFile(Filename: KOLstring): boolean;
    procedure LoadFromString(Str: KOLstring);
    procedure SaveToStream(Stream: PStream);
    procedure SaveToFile(Filename: KOLstring);
    procedure Include(P: PXMLParser);
    property OnLoad: TXMLSaveLoad read FOnLoad write FOnLoad;
    property OnSave: TXMLSaveLoad read FOnSave write FOnSave;
    constructor Create;
    destructor Destroy; virtual;
  end;

implementation

         {####################################
          #  General purpose constants       #
          #                  and  functions  #
          ####################################}
const
  START_BUF_SIZE = $100;
  ID_ID : KOLString = 'ID';

  UTF8_BOM     : array [1..3] of byte = ($EF, $BB, $BF);
  UTF16LE_BOM  : WideChar   = #$FEFF;
  UTF16BE_BOM  : WideChar   = #$FFFE;
  
  CR    : KOLChar = {$IFDEF UNICODE_CTRLS}#$000D{$ELSE}#$0D{$ENDIF};
  LF    : KOLChar = {$IFDEF UNICODE_CTRLS}#$000A{$ELSE}#$0A{$ENDIF};
  SP    : KOLChar = {$IFDEF UNICODE_CTRLS}#$0020{$ELSE}#$20{$ENDIF};
  TB    : KOLChar = {$IFDEF UNICODE_CTRLS}#$0009{$ELSE}#$09{$ENDIF};
  SPS   : KOLString = #$9; //{$IFDEF UNICODE_CTRLS}#$0020#$0020{$ELSE}#$20#$20{$ENDIF};
  CRLF  : KOLString = {$IFDEF UNICODE_CTRLS}#$000D#$000A{$ELSE}#$0D#$0A{$ENDIF};
  LTAN  : KOLChar = {$IFDEF UNICODE_CTRLS}#$003C{$ELSE}#$3C{$ENDIF};
  RTAN  : KOLChar = {$IFDEF UNICODE_CTRLS}#$003E{$ELSE}#$3E{$ENDIF};


  CR_V  : {$IFDEF UNICODE_CTRLS}WORD{$ELSE}BYTE{$ENDIF} = $D;
  LF_V  : {$IFDEF UNICODE_CTRLS}WORD{$ELSE}BYTE{$ENDIF} = $A;
  SP_V  : {$IFDEF UNICODE_CTRLS}WORD{$ELSE}BYTE{$ENDIF} = $20;
  TB_V  : {$IFDEF UNICODE_CTRLS}WORD{$ELSE}BYTE{$ENDIF} = $9;
  LTAN_V  : {$IFDEF UNICODE_CTRLS}WORD{$ELSE}BYTE{$ENDIF} = $3C;
  RTAN_V  : {$IFDEF UNICODE_CTRLS}WORD{$ELSE}BYTE{$ENDIF} = $3E;


function CheckEncoding(const Buf: PBytes): TEncoding;
begin
  Result := encANSI;
  case Buf^[0] of
    $FE: if Buf^[1] = $FF then Result := encUCS2BE;
    $FF: if Buf^[1] = $FE then Result := encUCS2LE;
    $EF: if (Buf^[1] = $BB) and (Buf^[2] = $BF) then Result := encUTF8;
    end;
end;

procedure Drop_BOM_UCS2LE(var WS: PWideChar);
begin
  if WS^ = UTF16LE_BOM
    then inc(WS);
end;

procedure Drop_BOM_UCS2BE(var WS: PWideChar);
begin
  if WS^ = UTF16BE_BOM
    then inc(WS);
end;

procedure Drop_BOM_UTF8(var S: PAnsiChar);
begin
  if
    (Ord(S[0]) = UTF8_BOM[1]) and
    (Ord(S[1]) = UTF8_BOM[2]) and
    (Ord(S[2]) = UTF8_BOM[3])
  then inc(S, 3);
end;

function Wide2Ansi(const WS: WideString; CP: cardinal = 1251): AnsiString;
var
  BufLen: integer;
begin
  Result := '';
  BufLen := WideCharToMultiByte(CP,0,@WS[1],-1,nil,0,nil,nil);
  if BufLen < 1 then exit;
  SetLength(Result, BufLen - 1);
  WideCharToMultiByte(CP,0,@WS[1],-1,@Result[1],BufLen,nil,nil);
end;

function Ansi2Wide(const S: AnsiString; CP: cardinal = 1251): WideString;
var
  BufLen: integer;
begin
  Result := '';
  BufLen := MultiByteToWideChar(CP, 0, @S[1], -1, nil, 0);
  if BufLen < 1 then exit;
  SetLength(Result, BufLen - 1{shl 2});
  MultiByteToWideChar(CP, 0, @S[1], -1, @Result[1], BufLen);
end;

function StrCode(const S: KOLString): KOLstring;
begin
    Result := S;
    try
    while KOLStrReplace(Result,'&','@@amp;') do;
    while KOLStrReplace(Result,'@@amp;','&amp;') do;
    while KOLStrReplace(Result,'<','&lt;') do;
    while KOLStrReplace(Result,'>','&gt;') do;
    while KOLStrReplace(Result,'''','&apos;') do;
    while KOLStrReplace(Result,'"','&quot;') do;
    {$IFDEF REPLACE_1310}
    while KOLStrReplace(Result,CRLF,'&crlf;') do;
    {$ENDIF}
    except
      asm
        NOP
      end;
    end;
end;

function StrDecode(const S: KOLString): KOLstring;
begin
    Result := S;
    try
    while KOLStrReplace(Result,'&amp;','&') do;
    while KOLStrReplace(Result,'&lt;','<') do;
    while KOLStrReplace(Result,'&gt;','>') do;
    while KOLStrReplace(Result,'&apos;','''') do;
    while KOLStrReplace(Result,'&quot;','"') do;
    {$IFDEF REPLACE_1310}
    while KOLStrReplace(Result,'&crlf;',CRLF) do;
    {$ENDIF}
    except
      asm
        NOP
      end;
    end;
end;

function ParseToken(const Body: KOLString; var Atr: KOLString): KOLString;
var
  P: PKOLChar;
  i: integer;
  N, A : PCharBuffer;
begin
  i := Length(Body);
  if i = 0 then exit;
  P := @Body[1];
  New(N, Create);
  New(A, Create);
  while ((P^ = SP) or (P^ = '<') or (P^ = '/')) and (i > 0) do
    begin
    inc(P);
    dec(i);
    end;
  while (not ((P^ = SP) or (P^ = '>') or (P^ = '/'))) and (i > 0) do
    begin
    N.PutChar(P^);
    inc(P);
    dec(i);
    end;
  while (not((P^ = '>') or (P^ = '/'))) and (i > 0) do
    begin
    A.PutChar(P^);
    inc(P);
    dec(i);
    end;
  Result := Trim (N.GetStr);
  N.Destroy;
  Atr := Trim(A.GetStr);
  A.Destroy;
end;
// ========================================================

         {####################################
          #        CharBuffer Object        #
          ####################################}

constructor TCharBuffer.Create;
begin
  Clear;
  //SetLength(FBuf, START_BUF_SIZE);
  
  //FSize := START_BUF_SIZE;
  //Clear;
  FPos := 0;
end;

destructor TCharBuffer.Destroy;
begin
  FPos := 0;
  FSize := 0;
  SetLength(FBuf, 0);
  FreeMem(@self);
end;

procedure TCharBuffer.PutStr(const S: KOLString);
var
  l: cardinal;
begin
  l := Length(S);
  if l = 0 then exit;
  FAlloc (l);
  {$IFDEF UNICODE_CTRLS}WStrCopy{$ELSE}StrCopy{$ENDIF}(@FBuf[FPos], @S[1]);
  inc(Fpos, l);
end;

procedure TCharBuffer.FAlloc(l: cardinal);
begin
  while (FPos + l ) >= FSize do begin
    FSize := FSize + (FSize div 4);
    SetLength(FBuf, FSize);
    end;
  Memory := @FBuf[0];
end;

procedure TCharBuffer.PutChar(const Ch: KOLChar);
begin
  FAlloc(1);
  FBuf[FPos] := Ch;
  inc(FPos);
end;

procedure TCharBuffer.PutBuf(const P: PKOLChar);
var
  l : cardinal;
begin
  l := {$IFDEF UNICODE_CTRLS}WStrLen{$ELSE}StrLen{$ENDIF}(P);
  FAlloc(l);
  {$IFDEF UNICODE_CTRLS}WStrCopy{$ELSE}StrCopy{$ENDIF}(@FBuf[FPos], P);
  inc(Fpos, l);
end;

procedure TCharBuffer.WrtStr(var S: KOLString);
begin
  S := '';
  if FPos = 0 then exit;
  SetLength(S, FPos);
  {$IFDEF UNICODE_CTRLS}WStrCopy{$ELSE}StrCopy{$ENDIF}(@S[1], @FBuf[0]);
end;

procedure TCharBuffer.WrtStream(P: PStream; CP: TEncoding);
var
  //Len, BufLen: integer;
  AnsiBuf : AnsiString;
  WideBuf : WideString;
  //Utf8Buf  : UTF8String;
begin
case CP of
    encUCS2LE, encUCS2BE: begin
        {$IFDEF UNICODE_CTRLS}
        P.Write(Memory^, FPos shl 1);
        {$ELSE}
        AnsiBuf := GetStr;
        WideBuf := Ansi2Wide(AnsiBuf);
        P.Write(WideBuf[1], Length(WideBuf) shl 1);
        WideBuf := '';
        {$ENDIF}
        end;
    encUTF8: begin
        {$IFDEF UNICODE_CTRLS}
        WideBuf := GetStr;
        AnsiBuf := Wide2Ansi(WideBuf, 65001);
        P.Write(AnsiBuf[1], Length(AnsiBuf));
        {$ELSE}
        AnsiBuf := GetStr;
        WideBuf := Ansi2Wide(AnsiBuf, 65001);
        AnsiBuf := Wide2Ansi(WideBuf, 65001);
        P.Write(AnsiBuf[1], Length(AnsiBuf));
        {$ENDIF}
        end;
    ELSE begin
        {$IFDEF UNICODE_CTRLS}
        WideBuf := GetStr;
        AnsiBuf := Wide2Ansi(WideBuf);
        P.Write(AnsiBuf[1], Length(AnsiBuf));
        {$ELSE}
        P.Write(Memory^, FPos);
        {$ENDIF}
        end;
    end;
AnsiBuf := ''; WideBuf := '';
end;

function TCharBuffer.GetStr: KOLString;
begin
  Result := '';
  if FPos = 0 then exit;
  SetLength(Result, FPos);
  {$IFDEF UNICODE_CTRLS}WStrCopy{$ELSE}StrCopy{$ENDIF}(@Result[1], @FBuf[0]);
end;

function TCharBuffer.FGetSize: cardinal;
begin
  Result := FPos{$IFDEF UNICODE_CTRLS}* SizeOf(KOLChar){$ENDIF};
end;

procedure TCharBuffer.Clear;
begin
  FSize := START_BUF_SIZE;
  SetLength(FBuf, FSize);
  FillChar(FBuf[0], FSize, 0);
  FPos := 0;
  Memory := @FBuf[0];
end;
// ========================================================

         {####################################
          #         Atribute Object          #
          ####################################}

constructor TXMLAtributeList.Create;
begin
  inherited;
  FAtrList := NewList;
end;

destructor TXMLAtributeList.Destroy;
begin
  Clear;
  FAtrList.Free;
  FAtrList := nil;
  inherited;
end;

function TXMLAtributeList.FGetCount: integer;
begin
  Result := FAtrList.Count;
end;

procedure TXMLAtributeList.Clear;
var
  i, y: integer;
begin
  y := Count - 1;
  for i := 0 to y do
	Dispose( PXMLAtribute(FAtrList.Items[i]));
  FAtrList.Clear;
end;

procedure TXMLAtributeList.Remove(index: integer);
begin
  if (index >= Count) or (index < 0) then exit;
  Dispose(PXMLAtribute(FAtrList.Items[index]));
  FAtrList.Delete(index);
end;

procedure TXMLAtributeList.Move(CurIdx, NewIdx: integer);
begin
  FAtrList.MoveItem(CurIdx, NewIdx);
end;

function TXMLAtributeList.IndexByName(AtrName: KOLString): integer;
var
  i, y: integer;
  a, b: KOLString;
begin
  Result := -1;
  y := Count;
  i := 0;
  a := {$IFDEF UNICODE_CTRLS}WUpperCase{$ELSE}UpperCase{$ENDIF}(AtrName);
  while  (i < y) and (Result = -1) do begin
	b := {$IFDEF UNICODE_CTRLS}WUpperCase{$ELSE}UpperCase{$ENDIF}(AtributeName[i]);
	if a = b then Result := i;
	inc(i);
	end;
end;

function TXMLAtributeList.FGetAtributeByIndex(index: integer): PXMLAtribute;
begin
  Result := nil;
  if (index >= Count) or  (index < 0) then exit;
  Result := FAtrList.Items[index];
end;

function TXMLAtributeList.FGetAtributeName(index: integer): KOLString;
begin
  Result := AtributeByIndex[index].AT_Name;
end;

procedure TXMLAtributeList.FSetAtributeName(index: integer; const value: KOLString);
begin
  AtributeByIndex[index].AT_Name := value;
end;

function TXMLAtributeList.FGetAtributeValue(index: integer): KOLString;
begin
  Result := AtributeByIndex[index].AT_Value;
end;

procedure TXMLAtributeList.FSetAtributeValue(index: integer; const value: KOLString);
begin
  AtributeByIndex[index].AT_Value := value;
end;

function TXMLAtributeList.FGetAtribute(AtrName: KOLString): KOLString;
var
  i : integer;
begin
  Result := '';
  i := IndexByName(AtrName);
  if i = -1 then exit;
  Result := AtributeValue[i];
end;

procedure TXMLAtributeList.FSetAtribute(AtrName: KOLString; const value: KOLString);
var
  i: integer;
  p: PXMLAtribute;
begin
  i := IndexByName(AtrName);
  if i = -1
   then begin
	New(p);
	p.AT_Name := AtrName;
	p.AT_Value := value;
	FAtrList.Add(p);
	end
   else AtributeValue[i] := value;
end;

{$IFDEF Atr_EX}

         //       READ        //
 
function TXMLAtributeList.FGetAsInteger(AtrName: KOLString): integer;
begin
  Result := Str2Int(Trim(Atribute[AtrName]));
end;
 
function TXMLAtributeList.FGetAsBoolean(AtrName: KOLString): boolean;
var 
  a: KOLstring;
begin
  a := {$IFDEF UNICODE_CTRLS}WUpperCase{$ELSE}UpperCase{$ENDIF}(Trim(Atribute[AtrName]));
  if (a = 'TRUE') or (a = 'ENABLED') or (a = '1')
	then Result := true
	else Result := false;
end;
 
function TXMLAtributeList.FGetAsFloat(AtrName: KOLString): extended;
begin
  Result := Str2Extended(Trim(Atribute[AtrName]));
end;
 
function TXMLAtributeList.FGetAsDateTime(AtrName: KOLString): TDateTime;
begin
	Result := Str2DateTimeShort(Atribute[AtrName]);
end;

         //       WRITE       //
 
procedure TXMLAtributeList.FSetAsInteger(AtrName: KOLString; const value: integer);
begin
  Atribute[AtrName] := Int2Str(value);
end;

procedure TXMLAtributeList.FSetAsBoolean(AtrName: KOLString; const value: boolean);
begin
  if value
	then Atribute[AtrName] := 'True'
	else Atribute[AtrName] := 'False';
end;

procedure TXMLAtributeList.FSetAsFloat(AtrName: KOLString; const value: extended);
begin
  Atribute[AtrName] := Extended2Str(value);
end;

procedure TXMLAtributeList.FSetAsDateTime(AtrName: KOLString; const value: TDateTime);
begin
  Atribute[AtrName] := DateTime2StrShort(value);
end;
{$ENDIF}

function TXMLAtributeList.FGetLine: KOLString;
var 
  i, y: integer;
  Buf: PCharBuffer;
begin
  Result := '';
  y := Count;
  if y = 0 then exit;
  New(Buf, Create);
  i := 0;
  while (i < y) do begin
    Buf.PutStr(AtributeByIndex[i].AT_Name);
    Buf.PutStr('="');
    Buf.PutStr(AtributeByIndex[i].AT_Value);
    Buf.PutStr('" ');
    inc(i);
	end;
  Result := TrimRight(Buf.GetStr);
  Buf.Destroy;
end;

procedure TXMLAtributeList.FSetLine(const value: KOLString);
var
  N, V: PCharBuffer;
  P: PKOLChar;
  L : integer;
begin
  Clear;
  L := Length(value);
  if L = 0 then exit;
  New(N, Create);
  New(V, Create);
  P := @value[1];
  while L > 0 do begin
    while (Ord(P^) in [SP_V, CR_V, LF_V, TB_V]) and (L > 0) do begin
      inc(P);
      dec(L);
      end;
    while (P^ <> '=') and ( L > 0) do begin
      N.PutChar(P^);
      inc(P);
      dec(L);
      end;
    if P^ = '=' then begin
      inc(P);
      dec(L);
      end;
    while (Ord(P^) in [SP_V, CR_V, LF_V, TB_V]) and (L > 0) do begin
      inc(P);
      dec(L);
      end;
    if P^ = '"' then begin
      inc(P);
      dec(L);
      end;
    while (P^ <> '"') and ( L > 0) do begin
      V.PutChar(P^);
      inc(P);
      dec(L);
      end;
    if P^ = '"' then begin
      inc(P);
      dec(L);
      end;
	  Atribute[TrimRight(N.GetStr)] := V.GetStr;
    N.Clear; V.Clear;
    end;
  N.Destroy;
  V.Destroy;
end;
// ========================================================

         {####################################
          #         Tokenizer Object         #
          ####################################}

procedure TXMLTokenizer.Clear;
begin
  FCnt := 0;
  FBuf := nil;
  FToken := TEmpty;
  FBBuf.Clear;
end;

procedure TXMLTokenizer.FSkipBlanks;
begin
  while (Ord(FBuf^) in [CR_V, LF_V, SP_V, TB_V]) and (FCnt > 0) do
    begin
    inc(FBuf);
    dec(FCnt);
    end;
end;

function TXMLTokenizer.Chk_Header: boolean;
var
  P: PKOLChar;
begin
  Result := false;
  P := FBBuf.Memory;
  if FBBuf.Count < 7 then exit;
  if(P[0] = '<') and
	  (P[1] = '?') and
	 ((P[2] = 'x') or (P[2] = 'X')) and
	 ((P[3] = 'm') or (P[3] = 'M')) and
	 ((P[4] = 'l') or (P[4] = 'L'))
	then Result := true;
end;

function TXMLTokenizer.Chk_Comment: boolean;
var
  P: PKOLChar;
begin
  P := FBBuf.Memory;
  if(P[0] = '<') and
    (P[1] = '!') and
    (P[2] = '-') and
    (P[3] = '-')
	 then Result := true
	 else Result := false; 
end;

function TXMLTokenizer.Chk_Close: boolean;
var
  //L: integer;
  P: PKOLChar;
  //ddd: boolean;
begin
  Result := false;
  P := FBBuf.Memory;
  //L := FBBuf.Count;
  if (P[0] = '<') and
     (P[1] = '/') {and
     (P[L-1] = '>')}
	 then Result := true;
	 //else {Result} ddd := true;
   //Result := not ddd;
end;

function TXMLTokenizer.Chk_OpenClose: boolean;
var
  L: integer;
  P: PKOLChar;
begin
  P := FBBuf.Memory;
  L := FBBuf.Count;
  if (P[0] 	 = '<') and
     (P[L-2] = '/') {and
     (P[L-1]   = '>')}
	 then Result := true
	 else Result := false;
end;

function TXMLTokenizer.Chk_CDATA: boolean;
var
  P: PKOLChar;
begin
  Result := false;
  P := FBBuf.Memory;
  if FBBuf.Count < 12 then exit;
  if(P[0] = '<') and
	  (P[1] = '!') and
	  (P[2] = '[') and
	 ((P[3] = 'C') or (P[3] = 'c')) and
	 ((P[4] = 'D') or (P[4] = 'd')) and
   ((P[5] = 'A') or (P[5] = 'a')) and
   ((P[6] = 'T') or (P[6] = 't')) and
   ((P[7] = 'A') or (P[7] = 'a')) and
    (P[8] = '[')
	then Result := true;
end;

function TXMLTokenizer.FReadCom: TTokenObj;
begin
  FBBuf.PutChar(FBuf^);
  while (not (Ord(FBuf^) in [0, RTAN_V])) and (FCnt > 0) do
    begin
    inc(FBuf);
    dec(FCnt);
    FBBuf.PutChar(FBuf^);
    end;
  if FBuf^ = #0 then begin
    Result := tEOF;
    exit;
    end;

  inc(FBuf);
  dec (FCnt);


  if Chk_Header
	  then Result := tHeader
  else if Chk_Comment
    then Result := FReadComment
  else if Chk_CDATA
    then Result := FReadCDATA
  else if Chk_Close
    then Result := tClose
  else if Chk_OpenClose
    then Result := tOpenClose
  else Result := tOpen;
end;

function TXMLTokenizer.FReadText: TTokenObj;
begin
  while (not (Ord(FBuf^) in [0, LTAN_V])) and (FCnt > 0) do
    begin
    FBBuf.PutChar(FBuf^);
    inc(FBuf);
    end;
  Result := tText;
end;

function TXMLTokenizer.FReadComment: TTokenObj;
var
  L: cardinal;
  P: PKOLChar;
begin
  P := FBBuf.Memory;
  L := FBBuf.Count;
  while (not ((P[L-1] = '>') and (P[L-2] = '-') and (P[L-3] = '-'))) and (FCnt > 0) do
    begin
    inc(FBuf);
    dec(FCnt);
    FBBuf.PutChar(FBuf^);
    end;
  if FBuf^ = #0
    then Result := tEOF
    else Result := tComment;
end;

function TXMLTokenizer.FReadCDATA: TTokenObj;
var
  L: cardinal;
  P: PKOLChar;
begin
  P := FBBuf.Memory;
  L := FBBuf.Count;
  while (not ((P[L-1] = '>') and (P[L-2] = ']') and (P[L-3] = ']'))) and (FCnt > 0) do
    begin
    inc(FBuf);
    dec(FCnt);
    FBBuf.PutChar(FBuf^);
    end;
  if FBuf^ = #0
    then Result := tEOF
    else Result := tComment;
end;

function TXMLTokenizer.NextToken: TTokenObj;
begin
  Result := tEOF;
  FToken := tEOF;
  FBody := '';
  FSkipBlanks;
  if (FBuf^ = #0) or (FCnt = 0) then exit;

  if FBuf^ = LTAN
    then Result := FReadCom
    else Result := FReadText;

  FToken := Result;
  FBody := FBBuf.GetStr;
  FBBuf.Clear;
end;

procedure TXMLTokenizer.Initilize(const Buf: PKOLChar; const Len: cardinal);
begin
  FBuf := Buf;
  FLen := Len;
  FCnt := FLen;
  FToken := tEmpty;
end;


constructor TXMLTokenizer.Create;
begin
  New (FBBuf, Create);
end;

destructor TXMLTokenizer.Destroy;
begin
  FBody := '';
  FBBuf.Destroy;
  FBBuf := nil;
  FCnt := 0;
  inherited;
end;
// ========================================================

         {####################################
          #         XML Node Object          #
          ####################################}
constructor TXMLNode.Create;
begin
  //inherited;
  New(FAtributes, Create);
  FNodeList := NewList;
end;

destructor TXMLNode.Destroy;
begin
  if FParent<> nil then FParent.FNodeList.Remove(@self);
  Clear;
  FAtributes.Free;
  FAtributes := nil;
  FNodeList.Free;
  FNodeList := nil;
  FParent := nil;
  FParser := nil;
  FRoot := nil;
  FTokenizer := nil;
  inherited;
end;

function TXMLNode.NewNode: PXMLNode;
begin
  New(Result, Create);
  Result.FRoot := FRoot;
  Result.FParent := @self;
  Result.FTokenizer := FTokenizer;
  Result.FParser := FParser;
  Result.FParentCount := FParentCount + 1;
  FNodeList.Add(Result);
end;

function TXMLNode.NodeIndexFrom(NodeName: KOLstring; IndexFrom: integer = 0): integer;
var
  i, y: integer;
begin
    Result:= -1;
    i:= IndexFrom;
    y := NodeCount - 1;
    while (i <= y) and (Result = -1) do begin
      if AnsiCompareText(Node[i].ND_Name, NodeName) = 0 then Result:= i;
      inc(i);
    end;
end;

function TXMLNode.FindNodeByAtribute(AtrName, Str: KOLstring; IndexFrom: integer = 0): PXMLNode;
var
  i, y: integer;
begin
  Result := nil;
  i := IndexFrom;
  y := NodeCount;
  while (Result = nil) and (i <= y) do
    begin
    if Node[i].Atributes.Atribute[AtrName] = Str then Result := Node[i];
    inc(i);
    end;
end;

procedure TXMLNode.MoveNode(CurIdx, NewIdx: integer);
begin
  FNodeList.MoveItem(CurIdx, NewIdx);
end;

procedure TXMLNode.RemoveNode(index: cardinal);
begin
  PXMLNode(FNodeList.Items[index]).Free;
end;

procedure TXMLNode.Clear;
begin
  while NodeCount > 0 do RemoveNode(0);
  FNDName := '';
  FText := '';
  Atributes.Clear;
  FAssigned := false;
end;

procedure TXMLNode.FSetName(const value: KOLString);
begin
  FNDName := value;
  FAssigned := true;
end;

procedure TXMLNode.FSetText(const value: KOLString);
begin
  FText := value;
  FAssigned := true;
end;

function TXMLNode.FGetXML: KOLString;
var
  P : PStream;
  CP: TEncoding;
begin
  CP := FParser.FEncoding;
  FParser.FEncoding := {$IFDEF UNICODE_CTRLS}encUCS2LE{$ELSE}encANSI{$ENDIF};
  P := NewMemoryStream;
  FSaveToStream(P);
  FParser.FEncoding := CP;
  P.Position := 0;
  Result := P.{$IFDEF UNICODE_CTRLS}ReadWStrZ{$ELSE}ReadStrZ{$ENDIF};
  P.Free;
end;

procedure TXMLNode.FSetXML(const value: KOLString);
var
  P : PKOLChar;
  L : cardinal;
begin
  Clear;
  P := @value[1];
  L := {$IFDEF UNICODE_CTRLS}WStrLen{$ELSE}StrLen{$ENDIF}(P);
  FTokenizer.Initilize(P, L);
  FLoadFromToken;
end;

function TXMLNode.FGetID: KOLString;
begin
  Result := Atributes.Atribute[ID_ID];
end;

procedure TXMLNode.FSetID(const value: KOLString);
begin
  Atributes.Atribute[ID_ID] := value;
end;

procedure TXMLNode.FSetParent(const value: PXMLNode);
begin
//
end;

function TXMLNode.FGetFirstNode: PXMLNode;
begin
  Result := nil;
  if NodeCount > 0
    then Result := Node[0]
    else if SafeAccess then Result := NewNode;
end;

function TXMLNode.FGetLastNode: PXMLNode;
var
  i: integer;
begin
  Result := nil;
  i := NodeCount - 1;
  if i > -1
    then Result := Node[i]
    else if SafeAccess then Result := NewNode;
end;

function TXMLNode.FGetChildCount: integer;
begin
  Result := FNodeList.Count;
end;

function TXMLNode.FGetNodeByIndex(index: integer): PXMLNode;
begin
  Result := FNodeList.Items[index];
end;

function TXMLNode.FGetNodeByName(NodeName: KOLString): PXMLNode;
var
  i: integer;
begin
  Result := nil;
  i := FGetNodeIndex(NodeName);
  if i > -1 
    then Result := Node[i]
    else if SafeAccess then begin
      Result := NewNode;
      Result.ND_Name := NodeName;
      end;
end;

function TXMLNode.FGetNodeIndex(NodeName: KOLString): integer;
begin
  Result := NodeIndexFrom(NodeName);
end;

function TXMLNode.FGetNodeById(ID_String: KOLString): PXMLNode;
var
  i, y: integer;
begin
  Result := nil;
  i := 0;
  y := NodeCount;
  while (Result <> nil) and ( i <= y) do
    begin
    if Node[i].ND_ID = ID_String then Result := Node[i];
    inc(i);
    end;
   if (Result = nil) and SafeAccess then begin
    Result := NewNode;
    Result.ND_ID := ID_String;
    end;
end;

function TXMLNode.FCheckHasAtributes: boolean;
begin
  if Atributes.Count = 0
    then Result := false
    else Result := true;
end;

function TXMLNode.FCheckBlanked: boolean;
begin
  if HasChild or HasText
    then Result := false
    else Result := true;
end;

function TXMLNode.FCheckChildrens: boolean;
begin
  if NodeCount = 0
    then Result := false
    else Result := true;
end;

function TXMLNode.FCheckText: boolean;
begin
  if Trim(FText) = ''
    then Result := false
    else Result := true;
end;

function TXMLNode.FCheckSafeAccess: boolean;
begin
  Result := true;
  if FParser = nil then exit;
  Result := FParser.FSafeAccess;
end;

function TXMLNode.FCheckFormated: boolean;
begin
  Result := true;
  if FParser = nil then exit;
  Result := FParser.FSafeAccess;
end;

function TXMLNode.FGetEncoding: TEncoding;
begin
  Result := encANSI;
  if FParser = nil then exit;
  Result := FParser.FEncoding;
end;

function TXMLNode.FGetTabulate: KOLstring;
begin
  Result := '';
  if FParentCount <= 0 then exit;
  Result := StrRepeat(SPS, FParentCount);
end;

function TXMLNode.FGetOpenToken: KOLstring;
var
  S: PCharBuffer;
begin
  New(S, Create);
  S.PutChar('<');
  S.PutStr(FNDName);
  if HasAtributes then begin
    S.PutChar(SP);
    S.PutStr(Atributes.Line);
    S.PutChar(SP);
    end;
  S.PutChar('>');
  Result := S.GetStr;
  S.Destroy;
end;

function TXMLNode.FGetOpenCloseToken: KOLstring;
var
  S: PCharBuffer;
begin
  New(S, Create);
  S.PutChar('<');
  S.PutStr(FNDName);
  if HasAtributes then begin
    S.PutChar(SP);
    S.PutStr(Atributes.Line);
    S.PutChar(SP);
    end;
  S.PutStr('/>');
  Result := S.GetStr;
  S.Destroy;
end;

function TXMLNode.FGetCloseToken: KOLstring;
begin
  Result := '</' + FNDName + '>';
end;

procedure TXMLNode.FSaveToStream(Stream: PStream);
var
  S: PCharBuffer;
  i, y: integer;
  Enc: TEncoding;
  Fmt: boolean;
begin
  Enc := FParser.FEncoding;
  Fmt := FParser.FFormated;
  New(S, Create);
  if not Blanked
    then
      begin
      if Fmt then S.PutStr(FGetTabulate);
      S.PutStr(FGetOpenToken);
      S.PutStr(StrCode(FText));

      S.WrtStream(Stream, Enc);
      S.Clear;

      if HasChild
        then
          begin
          y := NodeCount - 1;
          for i := 0 to y do
            begin
            S.PutStr(CRLF);
            S.WrtStream(Stream, Enc);
            S.Clear;
            Node[i].FSaveToStream(Stream);
            end;
          if Fmt then S.PutStr(CRLF + FGetTabulate);
          end;
      S.PutStr(FGetCloseToken);
      S.WrtStream(Stream, Enc);
      end
    else
      begin
      if Fmt then S.PutStr(FGetTabulate);
      S.PutStr(FGetOpenCloseToken);
      S.WrtStream(Stream, Enc);
      end;
    S.Destroy;
end;


procedure TXMLNode.FLoadFromToken;
var
  N, A: KOLString;
  X, Y: cardinal;
begin
  while True do begin
    N := ParseToken(FTokenizer.Body, A);
    //msgok(N + #13#10 + A + #13#10 + int2str(integer(FTokenizer.Token)));
    if Assigned( FParser ) then if Assigned( FParser.FOnLoad )
      then begin
      X := FTokenizer.FLen;
      Y := FTokenizer.FLen - FTokenizer.FCnt;
      FParser.FOnLoad(FParser, stgParse, FTokenizer.Token, @N[1], Y, X);
      end;
    case FTokenizer.Token of
      // ---------- end of buffer
      tEOF      : break;
      // ---------- <sample>
      tOpen     : begin
                if IsAssigned
                  then begin
                    NewNode.FLoadFromToken;
                    end
                  else begin
                    ND_Name := N;
                    Atributes.Line := A;
                    end;
                end;
      // ---------- </sample>
      tClose    : begin
                if IsAssigned
                  then if (AnsiCompareText(N, ND_Name) = 0)
                    then break
                    else NewNode.FLoadFromToken
                  else begin
                    ND_Name := N;
                    Atributes.Line := A;
                    break;
                    end;
                end;
      // ---------- <sample />
      tOpenClose: begin
                if IsAssigned
                  then NewNode.FLoadFromToken
                  else begin
                    ND_Name := N;
                    Atributes.Line := A;
                    break;
                    end;
                end;
      // ----------
      tHeader,
      tComment,
      tEmpty    : {Err} ;
      // ---------- [CDATA
      tCDATA    :  ;
      // ---------- sample
      tText     : ND_Text := StrDecode(FTokenizer.Body);
      end;
    FTokenizer.NextToken;
  end;
end;

procedure TXMLNode.Include ( P: PXMLNode);
var
  NSrc, NDst : PXMLNode;
  i, y, z    : integer;
begin
  y := P.NodeCount - 1;
  for i := 0 to y do begin
    NSrc := P.Node[i];
    z := NodeIndexFrom(NSrc.ND_Name);
    if z = -1
      then begin
        NDst := NewNode;
        NDst.ND_Name := NSrc.ND_Name;
        NDst.ND_Text := NSrc.ND_Text;
        NDst.Atributes.Line := NSrc.Atributes.Line;
        NDst.Include(NSrc);
        end
      else begin
        Node[z].Include(NSrc);
        end;

    end;

end;
// ========================================================

         {####################################
          #          Parser  Object          #
          ####################################}
constructor TXMLParser.Create;
begin
  //inherited;
  New(FRoot, Create);
  New(FHeader, Create);
  New(FToken, Create);
  FRoot.FTokenizer := FToken;
  FRoot.FParent := FRoot;
  FRoot.FParser := @self;
  FRoot.FRoot := nil;
  FHasHeader := false;
  FFormated := true;
end;

destructor TXMLParser.Destroy;
begin
  FRoot.Free;
  FRoot := nil;
  FToken.Free;
  FToken := nil;
  FHeader.Free;
  FHeader := nil;
  inherited;
end;

procedure TXMLParser.LoadFromStream(Stream: PStream);
var
  BufTmp  : TBytes;
  Buf     : KOLString;

  BufTmpA : PAnsiChar;
  BufTmpW : PWideChar;

  CP      : TEncoding;

  I       : integer;
  LN      : cardinal;
begin
  I := Stream.Size - Stream.Position;
  SetLength(BufTmp, I + 2);
  FillChar(BufTmp[0], I + 2, 0);
  if Assigned(FOnLoad) then FOnLoad(@self, stgRead, tEmpty, nil, 0, 0);
  Stream.Read(BufTmp[0], I);
  BufTmpA := @BufTmp[0];
  BufTmpW:= @BufTmp[0];
  CP := encANSI;
  {  }
  if Assigned(FOnLoad) then FOnLoad(@self, stgDecode, tEmpty, nil, 0, 0);
  case CheckEncoding(@BufTmp) of
    encUCS2LE: // UCS2 Little Endian
        begin
        Drop_BOM_UCS2LE(BufTmpW);
        CP := encUCS2LE;
        LN := lstrlenW(BufTmpW);
        SetLength(Buf, LN);
        {$IFDEF UNICODE_CTRLS}
        WStrCopy(@Buf[1], BufTmpW);
        {$ELSE}
        BufTmpA := @Buf[1];
        {Fuck IT?}
        WideCharToMultiByte(1251, 0, BufTmpW, -1, @Buf[1], LN, nil, nil);
        {$ENDIF}
        end;
    encUCS2BE: // UCS2 Big Endian
        begin
        Drop_BOM_UCS2BE(BufTmpW);
        CP := encUCS2BE;
        LN := lstrlenW(BufTmpW);
        SetLength(Buf, LN);
        {$IFDEF UNICODE_CTRLS}
        WStrCopy(@Buf[1], BufTmpW);
        {$ELSE}
        BufTmpA := @Buf[1];
        {Fuck IT?}
        WideCharToMultiByte(1251, 0, BufTmpW, -1, @Buf[1], LN, nil, nil);
        {$ENDIF}
        end;
    encUTF8: // UTF-8
        begin
        Drop_BOM_UTF8(BufTmpA);
        CP := encUTF8;
        {$IFDEF UNICODE_CTRLS}
        LN := lstrlenA(BufTmpA);
        SetLength(Buf, LN);
        LN := Utf8ToUnicode(@Buf[1], BufTmpA, LN);
        SetLength(Buf, LN);
        {$ELSE}
        Buf := Utf8ToAnsi(BufTmpA);
        {$ENDIF}
        end;
    encANSI:    //   UCS wo BOM  ,
        begin   //    - IsUnicodeText
        //LN := lstrlenA(BufTmpA);
        //SetLength(Buf, LN);
        {$IFDEF UNICODE_CTRLS}
        //BufTmpW := @Buf[1];
        {Fuck IT?}
        Buf := Ansi2Wide(AnsiString(BufTmpA))
        //MultiByteToWideChar(1251, 0, BufTmpA, -1, @Buf[1], LN);
        {$ELSE}
        LN := lstrlenA(BufTmpA);
        SetLength(Buf, LN);
        StrCopy(@Buf[1], BufTmpA);
        {$ENDIF}
        end;
    end;
  FEncoding := CP;
  SetLength(BufTmp, 0);
  LoadFromString(Buf);
  Buf := '';
end;

function TXMLParser.LoadFromFile(Filename: KOLstring): boolean;
var
  S: PStream;
begin
  Result := true;
  S := NewReadFileStream(Filename);
  try
    S.Position := 0;
    LoadFromStream(S);
    S.Free;
  except
    Result := false;
  end;
end;

procedure TXMLParser.LoadFromString(Str: KOLString);
var
  Buf   : PKOLChar;
  Len   : cardinal;
  N, A  : KOLString;
begin
  Root.Clear;
  Len := Length(Str);
  Buf := @Str[1];
  FToken.Initilize(Buf, Len);

  while True do begin
    FToken.NextToken;
    N := ParseToken(FToken.Body, A);
    if Assigned(FOnLoad) then FOnLoad(@self, stgParse, FToken.Token, @N[1], FToken.FCnt, FToken.FLen);
    case FToken.Token of
        // ---------- end of buffer
        tEOF    : break;
        // ---------- <sample>
        tOpen   : Root.FLoadFromToken;
        // ---------- <?xml .. ?>
        tHeader : begin
                // 'header';
                HasHeader := true;
                A := Trim(ExcludeTrailingChar(A, '?'));
                Header.Line := A;
                end;
        // ---------- <!-- .. -->
        tComment: begin
                // 'comment';
                end;
        tClose,
        tOpenClose,
        tEmpty,
        tCDATA,
        tText   : begin
                // Err
                end;
        end;
    end;
  FToken.Clear;
end;

procedure TXMLParser.SaveToStream(Stream: PStream);
var
  Buf: PCharBuffer;
  S: KOLString;
  u1, u2, u3: Pointer;
begin
  u1 := @UTF16LE_BOM;
  u2 := @UTF16BE_BOM;
  u3 := @UTF8_BOM;
  New(Buf, Create);
  case FEncoding of
    encUCS2LE:  Stream.Write(u1^, 2);
    encUCS2BE:  Stream.Write(u2^, 2);
    encUTF8:    Stream.Write(u3^, 3);
    end;

  if HasHeader then
    begin
    //<?xml version="1.0" encoding="Windows-1251"?>
    if Header.Atribute['version'] = '' then Header.Atribute['version'] := '1.0';
    if Header.Atribute['encoding'] = '' then begin
      case FEncoding of
        encUCS2LE, encUCS2BE: S := 'unicode';
        encUTF8 : S := 'UTF-8';
        else S := 'Windows-1251';
        end;
      Header.Atribute['encoding'] := S;
      end;
    Buf.PutStr('<?xml ');
    Buf.PutStr(Header.Line);
    Buf.PutStr(' ?>');
    if Formated then Buf.PutStr (CRLF);
    Buf.WrtStream(Stream, FEncoding);
    end;
  Buf.Destroy;
  FRoot.FSaveToStream(Stream);
end;

procedure TXMLParser.SaveToFile(Filename: KOLstring);
var
  P: PStream;
begin
  P := NewWriteFileStream(Filename);
  SaveToStream(P);
  P.Free;
end;

procedure TXMLParser.FSetEncoding(const value: TEncoding);
var
  i: integer;
begin
  i := Header.IndexByName('encoding');
  if i >= 0 then Header.Remove(i);
  FEncoding := value;
end;

procedure TXMLParser.Include(P: PXMLParser);
{var
  I, Y: integer;
  S: KOLString;}
begin
  {if P = nil then exit;
  Y := P.Root.NodeCount - 1;
  if Y < 0 then exit;
  for I := 0 to Y do begin
    S := P.Root.Node[I].ND_Name}
end;

end.
